import logging

from pydantic import BaseModel

from .dialect import Dialect
from .lib import run_sync

log = logging.getLogger('sqly.database')


class Database(BaseModel):
    connection_string: str = ':memory:'
    dialect: Dialect = Dialect.SQLITE

    @classmethod
    def connection_string_dialect(cls, connection_string):
        dialect = None
        if connection_string == ':memory:':
            dialect = Dialect.SQLITE

        if connection_string.startswith('postgres'):
            for dialect in [Dialect.ASYNCPG, Dialect.PSYCOPG2, Dialect.POSTGRES]:
                try:
                    dialect.load_adaptor()
                    break
                except Exception:
                    pass

        if connection_string.startswith('mysql'):
            dialect = Dialect.MYSQL

        if not dialect:
            dialect = Dialect.PYODBC

        return dialect

    def connect(self):
        mod = self.dialect.load_adaptor()
        if self.dialect.adaptor == 'sqlalchemy':
            connection = mod.create_engine(self.connection_string).connect()
        else:
            connection = run_sync(mod.connect(self.connection_string))

        return connection



from dataclasses import dataclass

from .dialect import DEFAULT_DIALECT
from .query import Query


@dataclass
class SQL:
    """
    Interface class: Enable creating and rendering queries with a given dialect. All
    queries are rendered with that dialect.
    """

    dialect: str = DEFAULT_DIALECT

    def query(self, query_text, data):
        return Query(query_text, dialect=self.dialect).render(data)

    def select(self, tablename, data, filter_keys=None):
        filter_data = {k: data[k] for k in filter_keys or data}
        return Query(
            [
                f'select {Query.fields(data)}',
                f'from {tablename}',
                f'where {Query.where(filter_data)}',
            ],
            dialect=self.dialect,
        ).render(data)

    def insert(self, tablename, data):
        return Query(
            [
                f'insert into {tablename}',
                f'({Query.fields(data)}) values ({Query.params(data)})',
                'returning *' if self.dialect.supports_returning else '',
            ],
            dialect=self.dialect,
        ).render(data)

    def update(self, tablename, data, filter_keys=None):
        filter_data = {k: data[k] for k in filter_keys or data}
        return Query(
            [
                f'update {tablename} set {Query.assigns(data)}',
                f'where {Query.where(filter_data)}',
                'returning *' if self.dialect.supports_returning else '',
            ],
            dialect=self.dialect,
        ).render(data)

    def delete(self, tablename, data, filter_keys=None):
        filter_data = {k: data[k] for k in filter_keys or data}
        return Query(
            [
                f'delete from {tablename} where {Query.where(filter_data)}',
                'returning *' if self.dialect.supports_returning else '',
            ],
            dialect=self.dialect,
        ).render(filter_data)

    def upsert(self, tablename, data, key_fields):
        return Query(
            [
                f'INSERT INTO {tablename} ({Query.fields(data)})',
                f'VALUES ({Query.params(data)})',
                f'ON CONFLICT ({key_fields}) DO UPDATE',
                f'SET {Query.assigns(data, excludes=key_fields)}',
                'RETURNING *' if self.dialect.supports_returning else '',
            ],
            dialect=self.dialect,
        ).render(data)



import json
import re
from enum import Enum
from importlib import import_module


class OutputFormats(Enum):
    COLON = ':field'
    PERCENT = '%(field)s'
    QMARK = '?'
    NUMBERED = '$i'


class Dialect(Enum):
    # the value of each Dialect represents the syntax that it uses for query parameters
    EMBEDDED = 'embedded'
    MYSQL = 'mysql'
    SQLITE = 'sqlite'
    POSTGRES = 'postgres'
    ASYNCPG = 'asyncpg'
    PSYCOPG2 = 'psycopg2'
    SQLALCHEMY = 'sqlalchemy'
    PYODBC = 'pyodbc'

    @property
    def supports_returning(self):
        return self in [self.POSTGRES, self.PSYCOPG2, self.ASYNCPG]

    @property
    def output_format(self):
        return {
            self.EMBEDDED: OutputFormats.COLON,
            self.SQLALCHEMY: OutputFormats.COLON,
            self.MYSQL: OutputFormats.PERCENT,
            self.PSYCOPG2: OutputFormats.PERCENT,
            self.SQLITE: OutputFormats.QMARK,
            self.POSTGRES: OutputFormats.NUMBERED,
            self.ASYNCPG: OutputFormats.NUMBERED,
            self.PYODBC: OutputFormats.QMARK,
        }[self]

    @property
    def adaptor(self):
        return {
            self.EMBEDDED: 'sqlalchemy',
            self.SQLALCHEMY: 'sqlalchemy',
            self.SQLITE: 'sqlite3',
            self.MYSQL: 'mysql',
            self.ASYNCPG: 'asyncpg',
            self.PSYCOPG2: 'psycopg2',
            self.POSTGRES: 'pyodbc',
        }[self]

    def load_adaptor(self):
        return import_module(self.adaptor)

    def render(self, query_string, data):
        """Render a query_string and its parameter values for this dialect.

        Returns 2-tuple: (query: str, data: dict)
        """
        pattern = r"(?<!\\):(\w+)\b"
        fields = []  # ordered list of fields for positional outputs (closure)

        def replace_parameter(match):
            field = match.group(1)

            # Build the ordered fields list for positional outputs (fields in closure)
            if (
                self.output_format in [OutputFormats.QMARK, OutputFormats.NUMBERED]
                or field not in fields
            ):
                fields.append(field)

            # Return the field formatted for the output type
            if self.output_format == OutputFormats.COLON:
                return f":{field}"
            elif self.output_format == OutputFormats.PERCENT:
                return f"%({field})s"
            elif self.output_format == OutputFormats.QMARK:
                return "?"
            elif self.output_format == OutputFormats.NUMBERED:
                return f"${len(fields)}"
            else:
                raise ValueError(
                    'Dialect %r output_format %r not supported'
                    % (self, self.output_format)
                )

        # 1. Escape string parameters in the PERCENT output format
        if self.output_format == OutputFormats.PERCENT:
            # any % must be intended as literal and must be doubled
            query_string = query_string.replace('%', '%%')

        # 2. Replace the parameter with its dialect-specific representation
        query_string = re.sub(pattern, replace_parameter, query_string).strip()

        # 3. Un-escape remaining escaped colon params
        if self.output_format == OutputFormats.COLON:
            # replace \:word with :word because the colon-escape is no longer needed.
            query_string = re.sub(r"\\:(\w+)\b", r":\1", query_string)

        # 4. Build the parameter_values dict or list for use with the query
        if self.output_format in [OutputFormats.COLON, OutputFormats.PERCENT]:
            # parameter_values is a dict of fields
            parameter_values = {
                key: json.dumps(val) if isinstance(val, dict) else val
                for key, val in {field: data[field] for field in fields}.items()
            }
        elif self.output_format in [OutputFormats.QMARK, OutputFormats.NUMBERED]:
            parameter_values = [
                json.dumps(val) if isinstance(val, dict) else val
                for val in [data[field] for field in fields]
            ]
        else:
            raise ValueError(
                'Dialect %r output_format %r not supported' % (self, self.output_format)
            )

        if self == Dialect.ASYNCPG:
            return tuple([query_string] + parameter_values)
        else:
            return query_string, parameter_values


DEFAULT_DIALECT = Dialect.EMBEDDED



import logging
from dataclasses import dataclass, field

from .dialect import DEFAULT_DIALECT
from .lib import walk_list

logger = logging.getLogger(__name__)


@dataclass
class Query:
    query: list = field(default_factory=list)
    dialect: str = DEFAULT_DIALECT

    def __post_init__(self):
        # allow the query to be a single string
        if isinstance(self.query, str):
            self.query = [self.query]

    def __str__(self):
        return ' '.join([str(q) for q in walk_list(self.query) if q]).strip()

    def render(self, data=None):
        """
        Render the query and its values for a given data input.
        """
        data = data or {}
        query_string = str(self)
        return self.dialect.render(query_string, data)

    @classmethod
    def fields(cls, data):
        return ', '.join(str(key) for key in data)

    @classmethod
    def params(cls, data):
        return ', '.join(f':{key}' for key in data)

    @classmethod
    def assigns(cls, data, excludes=None):
        excludes = excludes or []
        return ', '.join(f'{key}=:{key}' for key in data if key not in excludes)

    @classmethod
    def where(cls, data):
        return ' AND '.join(f'{key}=:{key}' for key in data)



import json
import os
import re
import uuid
from datetime import datetime, timezone
from glob import glob
from importlib import import_module
from pathlib import Path
from typing import List

import networkx as nx
import yaml
from pydantic import BaseModel, Field, validator

from .dialect import Dialect
from .lib import run_sync
from .settings import SQLY_UUID_NAMESPACE


def app_migrations_path(app):
    """
    For a given app name, get the path to the migrations directory.
    """
    mod = import_module(app)
    mod_filepath = Path(mod.__file__).parent
    return mod_filepath / 'migrations'


def make_migration_id():
    """
    An integer with the timestamp to millisecond resolution (17 digits => bigint)
    """
    return int(datetime.now(tz=timezone.utc).strftime('%Y%m%d%H%M%S%f')[:-3])


class Migration(BaseModel):
    id: int = Field(default_factory=make_migration_id)
    app: str
    name: str = Field(default='')
    depends: List[str] = Field(default_factory=list)
    applied: datetime = Field(default=None)
    doc: str = Field(default=None)
    up: str = Field(default=None)
    dn: str = Field(default=None)

    @validator('depends', pre=True, always=True)
    def depends_default_empty_list(cls, value):
        if isinstance(value, str):
            return json.loads(value)
        else:
            return value or []

    def __repr__(self):
        return (
            'Migration('
            + ', '.join(
                f'{key}={getattr(self, key)!r}' for key in ['key', 'depends', 'applied']
            )
            + ')'
        )

    def __str__(self):
        return self.yaml()

    def __hash__(self):
        return uuid.uuid3(SQLY_UUID_NAMESPACE, self.key).int

    @property
    def key(self):
        """
        The Migration key is used to uniquely identify the migration in the dependency
        graph. The key has the structure "{app}:{id}_{name}"

        - app = the name of the app that the Migration is in
        - id = the integer id for the Migration in that app
        """
        return f"{self.app}:{self.id}_{self.name or ''}"

    @property
    def filename(self):
        return f'{self.id}_{self.name or ""}.yaml'

    @classmethod
    def load(cls, filepath):
        with open(filepath) as f:
            data = yaml.safe_load(f.read())

        return Migration.parse_obj(data)

    @classmethod
    def key_load(cls, migration_key):
        return cls.load(cls.key_filepath(migration_key))

    @classmethod
    def key_filepath(cls, migration_key):
        app, basename = migration_key.split(':')
        return app_migrations_path(app) / f"{basename}.yaml"

    @classmethod
    def app_migrations(cls, app, include_depends=True):
        """
        For a given module name, get the existing migrations in that module. If
        include_depends is True (default), also include depends migrations from any app.
        """
        migration_filenames = glob(str(app_migrations_path(app) / '*.yaml'))
        migrations = set(cls.load(filename) for filename in migration_filenames)
        if include_depends is True:
            dms = set()
            for migration in migrations:
                dms |= migration.depends_migrations()

            migrations |= dms

        return migrations

    @classmethod
    def all_migrations(cls, *apps):
        migrations = set()
        for app in apps:
            migrations |= cls.app_migrations(app)

        return migrations

    @classmethod
    def create(cls, app, *other_apps, name=None):
        """
        Every new migration depends on all the "leaf" nodes in the existing migration
        graph. Leaf nodes are those with out_degree == 0 (no edges pointing out). See:
        <https://networkx.org/documentation/stable/reference/classes/generated/networkx.DiGraph.out_degree.html>.
        For a worked example, see:
        <https://stackoverflow.com/questions/31946253/find-end-nodes-leaf-nodes-in-radial-tree-networkx-graph/31953001>.
        """
        migrations = cls.all_migrations(app, *other_apps)
        graph = cls.graph(migrations)
        depends = [node for node in graph.nodes() if graph.out_degree(node) == 0]
        name = re.sub(r'\W+', '_', name or '')
        migration = cls(
            app=app, name=name or '', depends=depends, doc=None, up=None, dn=None
        )
        return migration

    @classmethod
    def graph(cls, migrations):
        """
        Given an iterable of Migrations, create a dependency graph of Migrations by key.
        """
        graph = nx.DiGraph()
        dag = {m.key: m.depends for m in migrations}
        for migration_key, migration_depends in dag.items():
            graph.add_node(migration_key)
            for depend in migration_depends:
                graph.add_edge(depend, migration_key)

        if not nx.is_directed_acyclic_graph(graph):
            raise nx.HasACycle(dag)

        return nx.transitive_reduction(graph)

    @classmethod
    def database_migrations(cls, database, connection=None):
        connection = connection or run_sync(database.connect())
        if database.dialect == Dialect.ASYNCPG:
            select = connection.fetch
        else:
            select = connection.execute

        try:
            rows = run_sync(select("select * from sqly_migration"))
            if database.dialect == Dialect.ASYNCPG:
                records = rows
            else:
                records = []
                fields = [d[0] for d in rows.description]
                for row in rows:
                    records.append(dict(zip(fields, row)))

            return {m.key: m for m in [cls(**record) for record in records]}

        except Exception as exc:
            print(exc)
            return {}

    @classmethod
    def migrate(cls, database, migration, connection=None, dryrun=False):
        """
        Migrate the database to this migration, either up or down, using the given
        (sqly) Database.

        1. Collate the list of applied migrations in the database with the list of
           migrations available in this application. (Edge case: There are no migrations
           in the database because the sqly_migrations table does not exist.)
        2. Calculate the graph path to reach this migration
            - if this migration has not been applied to the database, then the graph
              path is from the last applied predecessor up to this migration.
            - if this migration has been applied to the database, then the graph path is
              from the last applied successor down to this migration.
        3. Apply the sequence of migrations (either up or down).
        """
        connection = connection or run_sync(database.connect())

        db_migrations = cls.database_migrations(database, connection=connection)
        migrations = db_migrations | {
            m.key: m for m in cls.all_migrations(migration.app)
        }
        graph = cls.graph(migrations.values())

        if migration.key not in db_migrations:
            # apply 'up' migrations for all ancestors and this migration
            subgraph = nx.subgraph(graph, migration.ancestors(graph) | {migration.key})
            for key in nx.lexicographical_topological_sort(subgraph):
                if key not in db_migrations:
                    migrations[key].apply(
                        database, direction='up', connection=connection, dryrun=dryrun
                    )
        else:
            # apply 'dn' migrations for all descendants in reverse
            subgraph = nx.subgraph(graph, migration.descendants(graph))
            for key in reversed(list(nx.lexicographical_topological_sort(subgraph))):
                if key in db_migrations:
                    migrations[key].apply(
                        database, direction='dn', connection=connection, dryrun=dryrun
                    )

    def depends_migrations(self):
        """
        All migrations that this migration depends on, recursively.
        """
        dms = set()
        for depend in self.depends:
            depend_migration = self.load(self.key_filepath(depend))
            dms |= {depend_migration} | depend_migration.depends_migrations()
        return dms

    def ancestors(self, graph):
        return nx.ancestors(graph, self.key)

    def descendants(self, graph):
        return nx.descendants(graph, self.key)

    def yaml(self, **kwargs):
        return yaml.dump(self.dict(**kwargs), default_flow_style=False, sort_keys=False)

    def save(self):
        filepath = app_migrations_path(self.app) / self.filename
        os.makedirs(filepath.parent, exist_ok=True)
        with open(filepath, 'wb') as f:
            size = f.write(self.yaml().encode())

        return filepath, size

    def apply(self, database, direction='up', connection=None, dryrun=False):
        """
        Apply the migration (direction = 'up' or 'dn') to connection database. The
        entire migration script is wrapped in a transaction.
        """
        print(self.key, direction, end=' ... ')

        if dryrun:
            print('FAKE')
            return

        connection = connection or run_sync(database.connect())
        run_sync(connection.execute('begin;'))

        sql = getattr(self, direction, None)
        if sql:
            # (asyncpg does executescript via regular execute)
            if database.dialect == database.dialect.ASYNCPG:
                run_sync(connection.execute(sql))
            else:
                run_sync(connection.executescript(sql))

        if direction == 'up':
            query = self.insert_query(database)
        else:
            query = self.delete_query(database)

        run_sync(connection.execute(*query))
        run_sync(connection.execute('commit;'))
        print('OK')

    def insert_query(self, database):
        """
        Insert this migration into the sqly_migration table.
        """
        data = {k: v for k, v in self.dict(exclude_none=True).items()}
        if not isinstance(data.get('depends'), str):
            data['depends'] = json.dumps(data.get('depends') or [])
        keys = [k for k in data.keys()]
        params = [f':{k}' for k in keys]
        sql = f"""
            INSERT INTO sqly_migration ({','.join(keys)})
            VALUES ({','.join(params)});
            """
        return database.dialect.render(sql, data)

    def delete_query(self, database):
        sql = "DELETE FROM sqly_migration where id=:id"
        return database.dialect.render(sql, self.dict())



import asyncio
from importlib import import_module


def walk_list(a_list):
    """walk a nested list and yield items in a single stream"""
    for item in a_list:
        if not isinstance(item, str) and hasattr(item, '__iter__'):
            for i in flatten(item):
                yield i
        else:
            yield item


def flatten(a_list):
    """convert a walked list into a single list"""
    return list(walk_list(a_list))


def run_sync(result):
    """
    If result is a coroutine, evaluate it
    """
    if asyncio.iscoroutine(result):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(result)
    return result


def get_settings(mod_name, settings_mod_name=None):
    if not settings_mod_name:
        settings_mod_name = f"{mod_name}.settings"
    return import_module(settings_mod_name)


def get_logging_settings(settings, log=None):
    logging_settings = settings.LOGGING if hasattr(settings, 'LOGGING') else {}
    if log:
        logging_settings.update(level=int(log))
    return logging_settings



import os
import sys

import click
import networkx as nx

from .database import Database
from .migration import Migration


@click.group()
def main():
    pass


@main.command()
@click.option(
    '--name', required=False, help="A couple words describing the Migration's purpose"
)
@click.argument('app')
@click.argument('other_apps', nargs=-1)
def migration(app, other_apps, name):
    """
    Create a Migration in APP (importable python module) incorporating dependencies from
    OTHER_APPS
    """
    migration = Migration.create(app, *other_apps, name=name)
    migration.save()
    print(f"Created migration: {migration.key}")
    print("    depends:\n      -", '\n      - '.join(migration.depends or '[]'))


@main.command()
@click.argument('apps', nargs=-1)
@click.option('-i', '--include-depends', is_flag=True, help='Show dependencies')
def migrations(apps, include_depends=False):
    """
    List the Migrations in APPS
    """
    for app in apps:
        app_migrations = {
            m.key: m
            for m in Migration.app_migrations(app, include_depends=include_depends)
        }
        graph = Migration.graph(Migration.app_migrations(app, include_depends=True))
        for key in nx.lexicographical_topological_sort(graph):
            if key in app_migrations:
                print(key)
                migration = app_migrations[key]
                if migration.depends and include_depends:
                    print('\t=> ' + ', '.join(migration.depends))


@main.command()
@click.argument('migration_key')
@click.option(
    '-d',
    '--database-url',
    required=False,
    help="Datebase to migrate; default = env $DATABASE_URL",
)
@click.option(
    '-r',
    '--dryrun',
    is_flag=True,
    help="If present, shows but does not run the migrations",
)
def migrate(migration_key, database_url=None, dryrun=False):
    """
    Migrate database_url to the given MIGRATION_KEY (up or dn).
    """
    database_url = database_url or os.getenv('DATABASE_URL')
    if not database_url:
        print('--database-url or env $DATABASE_URL must be set', file=sys.stderr)
        sys.exit(1)

    dialect = Database.connection_string_dialect(database_url)
    database = Database(connection_string=database_url, dialect=dialect)

    migration = Migration.key_load(migration_key)
    Migration.migrate(database, migration, dryrun=dryrun)


if __name__ == '__main__':
    main()
